"use strict";

exports.fiber = fiber;

var _posterus = require("./posterus");

function fiber(iter) {
    validate(iter, isIterator);
    var out = new _posterus.Future();
    var pending = void 0;
    function onDone(error, value) {
        if (error) out.settle(maybeToFuture(error)); else out.settle(undefined, maybeToFuture(value));
        if (pending) pending.deinit();
        pending = undefined;
        iter.return();
    }
    function iterNext(input) {
        try {
            onIterNext(iter.next(input));
        } catch (err) {
            onDone(err);
        }
    }
    function iterThrow(error) {
        try {
            onIterNext(iter.throw(error));
        } catch (err) {
            onDone(err);
        }
    }
    function onIterNext(_ref) {
        var value = _ref.value, done = _ref.done;
        if (done) {
            onDone(undefined, value);
            return;
        }
        value = maybeToFuture(value);
        if (_posterus.isFuture(value)) {
            pending = value.map(onPendingDone);
            return;
        }
        iterNext(value);
    }
    function onPendingDone(error, value) {
        pending = undefined;
        if (error) iterThrow(error); else iterNext(value);
    }
    iterNext();
    return out.finally(onDone);
}

function maybeToFuture(value) {
    return _posterus.isFuture(value) ? value : isIterator(value) ? fiber(value) : isPromise(value) ? _posterus.Future.fromPromise(value) : value;
}

function isIterator(value) {
    return isObject(value) && isFunction(value.next) && isFunction(value.return) && isFunction(value.throw);
}

function isPromise(value) {
    return isObject(value) && isFunction(value.then) && isFunction(value.catch);
}

function isFunction(value) {
    return typeof value === "function";
}

function isObject(value) {
    return value != null && typeof value === "object";
}

function validate(value, test) {
    if (!test(value)) throw Error("Expected " + value + " to satisfy test " + test.name);
}